import torch
import numpy as np

from .utils import pytorch_input_adapter
from .utils import find_random_roll_values_for_tensor
from .utils import roll_torch_tensor
from .utils import get_random_rotation_angle
from .utils import rotate_image_tensor
from .utils import CascadeGaussianSmoothing

from .constants import UPPER_IMAGE_BOUND
from .constants import LOWER_IMAGE_BOUND
from .constants import UPPER_IMAGE_BOUND_GRAY
from .constants import LOWER_IMAGE_BOUND_GRAY

from .dreamer_utils import get_gradients

def dream_on_octave_with_masks(model, image_np, layers, iterations, lr,  custom_funcs = [None], max_rotation = 0.2, gradient_smoothing_coeff = None, gradient_smoothing_kernel_size = None, grad_mask =None, device = None, default_func = None):

    """
    Core function for image optimization with gradient masks 

    input args{
        model: pytorch model 
        image_np: 3 channel numpy image with shape (C,H,W)
        layers: list of layers whose outputs are to be used [model.layer1, model.layer2]
        lr: learning rate
        custom_funcs: list of custom functions to be applied with the corrresponding gradient masks 
        max_rotation: caps the max amount of rotation on the input tensor, helps reduce noise 
        gradient_smoothing_coeff: helps reduce fig frequency noise in gradients before adding
        gradient_smoothing_kernel_size: kernel size to be used while smoothing 
        grad_mask: list of gradient masks to be applied , make sure that len(grad_mask) = len(custom_funcs)
        device: cuda or CPU depending on the availability
        default_func: default func to be used if no custom_func is not given

    }
    """
        
    image_tensor = pytorch_input_adapter(image_np, device = device)
    if grad_mask is not None:
        grad_mask_tensors = [pytorch_input_adapter(g_mask, device = device).double() for g_mask in grad_mask]

    for i in range(iterations):
        """
        rolling 
        """
        roll_x, roll_y = find_random_roll_values_for_tensor(image_tensor)

        image_tensor_rolled = roll_torch_tensor(image_tensor, roll_x, roll_y) 
        
        """
        rotating
        """
        theta = get_random_rotation_angle(theta_max= max_rotation)

        image_tensor_rolled_rotated = rotate_image_tensor(image_tensor = image_tensor_rolled, theta = theta, device = device)

        """
        getting gradients
        """

        gradients_tensors = []
        for c in range(len(custom_funcs)):

            gradients_tensor = get_gradients(net_in = image_tensor_rolled_rotated, net = model, layers = layers,default_func = default_func,custom_func= custom_funcs[c]).detach()
            gradients_tensors.append(gradients_tensor)
        """
        unrotate and unroll gradients of the image tensor
        """
        gradients_tensors_unrotated  = [rotate_image_tensor(g, theta = -theta, device = device) for g in gradients_tensors]
        gradients_tensors = [roll_torch_tensor(g, -roll_x, -roll_y)  for g in gradients_tensors_unrotated]

        """
        image update
        """
        # print(gradient_smoothing_sigma, gradient_smoothing_kernel_size)
        
        if gradient_smoothing_kernel_size is not None and gradient_smoothing_coeff is not None:
            
            sigma = ((i + 1) / iterations) * 2.0 + gradient_smoothing_coeff
            gradients_tensors = [CascadeGaussianSmoothing(kernel_size = gradient_smoothing_kernel_size, sigma = sigma, device = device)(gradients_tensor.unsqueeze(0)).squeeze(0) for gradients_tensor in gradients_tensors]

            for m in range(len(gradients_tensors)):

                gradients_tensor = gradients_tensors[m]
                g_norm = torch.std(gradients_tensor)
                image_tensor.data = image_tensor.data + (lr *(gradients_tensor.data /g_norm) * grad_mask_tensors[m] )## can confirm this is still on the GPU if you have one
        
        else:
           for m in range(len(gradients_tensors)):

                gradients_tensor = gradients_tensors[m]
                g_norm = torch.std(gradients_tensor)
                
                image_tensor.data = image_tensor.data + (lr *(gradients_tensor.data /g_norm) * grad_mask_tensors[m] )## can confirm this is still on the GPU if you have one
        
        image_tensor.data = torch.max(torch.min(image_tensor.data.float(), UPPER_IMAGE_BOUND), LOWER_IMAGE_BOUND).squeeze(0)

    img_out = image_tensor.detach().cpu()

    img_out_np = img_out.numpy()
    img_out_np = img_out_np.transpose(1,2,0)
    
    return img_out_np

def dream_on_octave(model, image_np, layers, iterations, lr,  custom_func = None, max_rotation = 0.2, gradient_smoothing_coeff = None, gradient_smoothing_kernel_size = None, device = None, default_func = None):

    """
    Core function for image optimization with gradient masks 

    input args{
        model: pytorch model 
        image_np: 3 channel numpy image with shape (C,H,W)
        layers: list of layers whose outputs are to be used [model.layer1, model.layer2]
        lr: learning rate
        custom_func: custom loss defined by the user 
        max_rotation: caps the max amount of rotation on the input tensor, helps reduce noise 
        gradient_smoothing_coeff: helps reduce fig frequency noise in gradients before adding
        gradient_smoothing_kernel_size: kernel size to be used while smoothing 
        device: cuda or CPU depending on the availability
        default_func: default func to be used if no custom_func is not given
    }
    """

    image_tensor = pytorch_input_adapter(image_np, device = device)
    for i in range(iterations):
        """
        rolling 
        """

        roll_x, roll_y = find_random_roll_values_for_tensor(image_tensor)

        image_tensor_rolled = roll_torch_tensor(image_tensor, roll_x, roll_y) 
        
        """
        rotating
        """
        
        theta = get_random_rotation_angle(theta_max= max_rotation)


        image_tensor_rolled_rotated = rotate_image_tensor(image_tensor = image_tensor_rolled, theta = theta, device = device)

        """
        getting gradients
        """
        gradients_tensor = get_gradients(net_in = image_tensor_rolled_rotated, net = model, layers = layers, default_func = default_func ,custom_func= custom_func).detach()

        """
        unrotate and unroll gradients of the image tensor
        """
        gradients_tensor_unrotated  = rotate_image_tensor(gradients_tensor, theta = -theta, device = device)
        gradients_tensor = roll_torch_tensor(gradients_tensor_unrotated, -roll_x, -roll_y)  

        """
        image update
        """
        
        if gradient_smoothing_kernel_size is not None and gradient_smoothing_coeff is not None:
            
            sigma = ((i + 1) / iterations) * 2.0 + gradient_smoothing_coeff

            smooth_gradients_tensor = CascadeGaussianSmoothing(kernel_size = gradient_smoothing_kernel_size, sigma = sigma, device = device)(gradients_tensor.unsqueeze(0)).squeeze(0)
            g_norm = torch.std(smooth_gradients_tensor)

            image_tensor.data = image_tensor.data + lr * (smooth_gradients_tensor.data / g_norm) ## can confirm this is still on the GPU if you have one
        else:
            g_norm = torch.std(gradients_tensor)
            image_tensor.data = image_tensor.data + lr *(gradients_tensor.data /g_norm) ## can confirm this is still on the GPU if you have one

        image_tensor.data = torch.max(torch.min(image_tensor.data, UPPER_IMAGE_BOUND), LOWER_IMAGE_BOUND).squeeze(0)
       
    img_out = image_tensor.detach().cpu()

    img_out_np = img_out.numpy()
    img_out_np = img_out_np.transpose(1,2,0)
    
    return img_out_np


import cv2
import tqdm
import torch
import numpy as np
from tqdm import tqdm

from .utils import load_image
from .utils import preprocess_numpy_img
from .utils import pytorch_input_adapter
from .utils import pytorch_output_adapter
from .utils import post_process_numpy_image

from .constants import default_config
from .dreamer_utils import default_func_mean
from .dreamer_utils import make_octave_sizes

from .octave_utils import dream_on_octave_with_masks
from .octave_utils import dream_on_octave

class dreamer():

    """
    Main class definition for torch-dreams:

    model = Any PyTorch deep-learning model
    device = "cuda" or "cpu" depending on GPU availability
    self.config = dictionary containing everything required check the readme (https://github.com/Mayukhdeb/torch-dreams#a-closer-look) for a better explanantion. 
    self.default_func = default loss to be used if no custom_func is defined 
    """

    def __init__(self, model):
        self.model = model
        self.model = self.model.eval()
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        # model moves to GPU if available
        self.model = self.model.to(self.device)
        self.config = default_config.copy()

        self.default_func = default_func_mean
        self.dream_on_octave = dream_on_octave
        self.dream_on_octave_with_masks = dream_on_octave_with_masks

        print("dreamer init on: ", self.device)

    def deep_dream(self, config):

        for key in list(config.keys()):
            self.config[key] = config[key]

        image_path = self.config["image_path"]        
        original_image = load_image(image_path, grayscale= False)
        image_np = preprocess_numpy_img(original_image, grayscale= False)

        original_size = image_np.shape[:-1]

        octave_sizes = make_octave_sizes(
            original_size=original_size, num_octaves= self.config["num_octaves"], octave_scale=self.config["octave_scale"])

        """
        partial source for the next few lines: 
        https://github.com/ProGamerGov/Protobuf-Dreamer/blob/bb9943411129127220c131793264c8b24a71a6c0/pb_dreamer.py#L105
        """
    
        img = original_image.copy()
        octaves = []  

        for size in octave_sizes[::-1]:
            hw = img.shape[1], img.shape[0]
            lo = cv2.resize(img, size)
            hi = img- cv2.resize(lo, hw)
            img = lo
            octaves.append(hi)

        count = 0
        for new_size in tqdm(octave_sizes):

            image_np = cv2.resize(image_np, new_size)

            if  count > 0:
                hi = octaves[-count]
                image_np += hi

            image_np = self.dream_on_octave(
                model=self.model,
                image_np = image_np,
                layers = self.config["layers"],
                iterations = self.config["iterations"],
                lr= self.config["lr"],
                custom_func = self.config["custom_func"],
                max_rotation = self.config["max_rotation"],
                gradient_smoothing_coeff = self.config["gradient_smoothing_coeff"],
                gradient_smoothing_kernel_size= self.config["gradient_smoothing_kernel_size"], 
                default_func=self.default_func, 
                device=self.device
            )
            count += 1

        image_np = post_process_numpy_image(image_np)
        return image_np


    def deep_dream_with_masks(self, config):

        for key in list(config.keys()):
            self.config[key] = config[key]

        image_path = self.config["image_path"]

        original_image = load_image(self.config["image_path"], grayscale= False)
        image_np = preprocess_numpy_img(original_image, grayscale= False)

        original_size = image_np.shape[:-1]

        octave_sizes = make_octave_sizes(
            original_size=original_size, num_octaves=self.config["num_octaves"], octave_scale=self.config["octave_scale"])

        img = original_image.copy()
        octaves = []  

        for size in octave_sizes[::-1]:
            hw = img.shape[1], img.shape[0]
            lo = cv2.resize(img, size)
            hi = img- cv2.resize(lo, hw)
            img = lo
            octaves.append(hi)

        count = 0

        for new_size in tqdm(octave_sizes):

            image_np = cv2.resize(image_np, new_size)

            if  count > 0:
                hi = octaves[-count]
                image_np += hi

            if self.config["grad_mask"] is not None:
                grad_mask = [cv2.resize(g, new_size) for g in self.config["grad_mask"]]

            image_np = self.dream_on_octave_with_masks(
                model=self.model, 
                image_np=image_np, 
                layers= self.config["layers"], 
                iterations= self.config["iterations"], 
                lr= self.config["lr"], 
                custom_funcs= self.config["custom_func"], 
                max_rotation= self.config["max_rotation"],
                gradient_smoothing_coeff= self.config["gradient_smoothing_coeff"], 
                gradient_smoothing_kernel_size= self.config["gradient_smoothing_kernel_size"], 
                grad_mask= grad_mask, 
                device=self.device, 
                default_func=self.default_func
            )
            count += 1

        image_np = post_process_numpy_image(image_np)
        return image_np




import numpy as np
import torch

device = "cuda" if torch.cuda.is_available() else 'cpu'


IMAGENET_MEAN_1 = np.array([0.485, 0.456, 0.406], dtype=np.float32)
IMAGENET_STD_1 = np.array([0.229, 0.224, 0.225], dtype=np.float32)


IMAGENET_MEAN_1_GRAY = np.array([0.449], dtype=np.float32)
IMAGENET_STD_1_GRAY = np.array([0.226], dtype=np.float32)

LOWER_IMAGE_BOUND = torch.tensor((-IMAGENET_MEAN_1 / IMAGENET_STD_1).reshape(1, -1, 1, 1)).to(device)
UPPER_IMAGE_BOUND = torch.tensor(((1 - IMAGENET_MEAN_1) / IMAGENET_STD_1).reshape(1, -1, 1, 1)).to(device)

"""
Even though gray is not required for now, I plan to add a grayscale mode someday for single input channel models
"""
LOWER_IMAGE_BOUND_GRAY = torch.tensor((-IMAGENET_MEAN_1_GRAY / IMAGENET_STD_1_GRAY).reshape(1, -1, 1, 1)).to(device)
UPPER_IMAGE_BOUND_GRAY = torch.tensor(((1 - IMAGENET_MEAN_1_GRAY) / IMAGENET_STD_1_GRAY).reshape(1, -1, 1, 1)).to(device)

default_config = {
    "image_path": None,
    "layers": None,
    "octave_scale": None,
    "num_octaves": None,
    "iterations":None,
    "lr": None,
    "custom_func": None,
    "max_rotation": 0.5,
    "gradient_smoothing_coeff": None,
    "gradient_smoothing_kernel_size": None,
    "grad_mask": None,
}